@@ -22,7 +22,7 @@ from functools import lru_cache as cache
 from functools import reduce
 from importlib import metadata
 from importlib.machinery import ModuleSpec, SourceFileLoader
-from importlib.metadata import Distribution, PackageNotFoundError, PackagePath
+from importlib.metadata import Distribution
 from itertools import chain, product, zip_longest
 from logging import getLogger
 from numbers import Complex, Integral, Number, Rational, Real
@@ -31,11 +31,10 @@ from shutil import rmtree
 from sqlite3 import Cursor
 from subprocess import CalledProcessError, run
 from types import ModuleType
-from typing import Any, Iterable, Optional, Protocol, TypeVar, Union, get_args, get_origin, runtime_checkable
+from typing import Any, Optional, Protocol, TypeVar, Union, get_args, get_origin, runtime_checkable
 from warnings import catch_warnings, filterwarnings, warn
 
 import furl
-import packaging
 import requests
 from beartype import beartype
 from furl import furl as URL
@@ -43,10 +42,10 @@ from icontract import ensure, require
 from packaging import tags
 from packaging.specifiers import SpecifierSet
 
-from use import Hash, InstallationError, Modes, UnexpectedHash, VersionWarning, config, sessionID
-from use.hash_alphabet import JACK_as_num, hexdigest_as_JACK, num_as_hexdigest
-from use.messages import UserMessage, _web_pebkac_no_hash, _web_pebkac_no_version_no_hash
-from use.pydantics import PyPI_Project, PyPI_Release, RegistryEntry, Version
+from use import Hash, InstallationError, Modes, UnexpectedHash, VersionWarning, config
+from use.hash_alphabet import hexdigest_as_JACK, num_as_hexdigest
+from use.messages import UserMessage, _web_pebkac_no_hash
+from use.pydantics import PyPI_Project, PyPI_Release, PyPI_URL, RegistryEntry, Version
 from use.tools import pipes
 
 log = getLogger(__name__)
@@ -111,7 +110,7 @@ def get_supported() -> frozenset[PlatformTag]:  # cov: exclude
     (*ahem* pip, package_resources, packaging.tags *cough*)
     whose modules are notoriously resource-hungry.
 
-    Returns a set containing all platform _platform_tags
+    Returns a set containing all platform
     supported on the current system.
     """
     get_supported = None
@@ -132,7 +131,11 @@ def get_supported() -> frozenset[PlatformTag]:  # cov: exclude
 
     items: list[PlatformTag] = [PlatformTag(platform=tag.platform) for tag in get_supported()]
 
-    items.extend(PlatformTag(platform=str(tag)) for tag in packaging.tags._platform_tags())
+    # yay backwards compatibility..
+    if hasattr(tags, "platform_tags"):
+        items.extend(PlatformTag(platform=str(tag)) for tag in tags.platform_tags())
+    else:
+        items.extend(PlatformTag(platform=str(tag)) for tag in tags._platform_tags())
 
     return frozenset(items)
 
@@ -264,18 +267,8 @@ def _pebkac_no_hash(
     hash_algo: Hash,
     **kwargs,
 ) -> RuntimeWarning:
-    project = _get_data_from_pypi(package_name=package_name, version=version)
-    releases = _get_releases(project)
-
-    if version not in (r.version for r in releases):
-        last_version = project.releases >> reversed >> iter >> next
-        return RuntimeWarning(Message.no_distribution_found(package_name, version, last_version))
-
-    project = _filter_by_version(releases, version=version)
-    filtered = _filter_by_platform(
-        releases,
-        tags=get_supported(),
-    )
+    releases = _get_releases_from_pypi(package_name=package_name, version=version)
+    filtered = _filter_by_platform(releases, tags=get_supported())
     ordered = _sort_releases(filtered)
 
     if not no_browser:
@@ -311,7 +304,9 @@ def _pebkac_no_version_no_hash(
     **kwargs,
 ) -> Exception:
     # let's try to make an educated guess and give a useful suggestion
-    proj = _get_data_from_pypi(package_name=package_name)
+    proj = _get_project_from_pypi(package_name=package_name)
+    if isinstance(proj, Exception):
+        return proj
     releases = _get_releases(proj)
 
     ordered = releases >> _filter_by_platform(tags=get_supported()) >> _sort_releases
@@ -476,20 +471,23 @@ def _auto_install(
         return ImportError(msg)
 
     # else: we have to download the package and install it
-    project = _get_data_from_pypi(package_name=package_name, version=version)
+
+    releases = _get_releases_from_pypi(package_name=package_name, version=version)
+    if isinstance(releases, Exception):
+        return releases
     # we *did* ask the user to give us hashes of artifacts that *should* work, so let's check for those.
     # We can't be sure which one of those hashes will work on this platform, so let's try all of them.
 
     for H in user_provided_hashes:
         log.info(f"Attempting auto-installation of <{num_as_hexdigest(H)}>...")
         url = next(
-            (URL(purl.url) for purl in project.urls if H == int(purl.digests[hash_algo.name], 16)),
+            (URL(R.url) for R in releases if H == int(R.digests[hash_algo.name], 16)),
             None,
         )
 
         if not url:
             return UnexpectedHash(
-                f"user provided {user_provided_hashes}, do not match any of the {len(project.urls)} possible artifacts"
+                f"user provided {user_provided_hashes}, do not match any of the {len(releases)} possible artifacts"
             )
 
         # got an url for an artifact with a hash given by the user, let's install it
@@ -599,12 +597,10 @@ def _download_artifact(*, artifact_path: Path, url: URL, hash_algo: Hash, hash_v
 
 @beartype
 def _is_pure_python_package(artifact_path: Path, meta: dict) -> bool:
-    for n, s in product(meta["names"], importlib.machinery.EXTENSION_SUFFIXES):
-        if n.endswith(s):
-            return False
-    if ".tar" in str(artifact_path):
-        return False
-    return True
+    return next(
+        (False for n, s in product(meta["names"], importlib.machinery.EXTENSION_SUFFIXES) if n.endswith(s)),
+        ".tar" not in str(artifact_path),
+    )
 
 
 @beartype
@@ -766,20 +762,37 @@ def _load_venv_entry(*, module_name: str, installation_path: Path) -> ModuleType
     raise ImportError(msg)
 
 
-@cache(maxsize=512)
-def _get_data_from_pypi(*, package_name: str, version: str = None) -> PyPI_Project:
-    if version:
-        json_url = f"https://pypi.org/pypi/{package_name}/{version}/json"
-    else:
-        json_url = f"https://pypi.org/pypi/{package_name}/json"
-    response = requests.get(json_url)
+@beartype
+def _get_project_from_pypi(*, package_name: str) -> PyPI_Project | Exception:
+    # let's check if package name is correct
+    url = f"https://pypi.org/pypi/{package_name}/json"
+    response = requests.get(url)
     if response.status_code == 404:
-        raise ImportError(UserMessage.pebkac_unsupported(package_name))
+        return ImportError(UserMessage.pebkac_unsupported(package_name))
     elif response.status_code != 200:
-        raise RuntimeWarning(UserMessage.web_error(json_url, response))
+        return RuntimeWarning(UserMessage.web_error(url, response))
     return PyPI_Project(**response.json())
 
 
+@beartype
+def _get_releases_from_pypi(*, package_name: str, version: Version) -> list[PyPI_Release] | Exception:
+    # let's check if package name is correct
+    url = f"https://pypi.org/pypi/{package_name}"
+    response = requests.get(url)
+    if response.status_code == 404:
+        return ImportError(UserMessage.pebkac_unsupported(package_name))
+    elif response.status_code != 200:
+        return RuntimeWarning(UserMessage.web_error(url, response))
+    # let's check if the version is a thing
+    url = f"https://pypi.org/pypi/{package_name}/{version}/json"
+    response = requests.get(url)
+    if response.status_code == 404:
+        return RuntimeWarning(UserMessage.bad_version_given(package_name, version))
+    # looks good, let's get the releases for this version
+    urls = response.json()["urls"]
+    return [PyPI_Release(**url, version=version) for url in urls]
+
+
 @beartype
 def _filter_by_platform(releases: list[PyPI_Release], *, tags: frozenset[PlatformTag]) -> list[PyPI_Release]:
     def compatible(info: PyPI_Release, include_sdist=False) -> bool:
@@ -790,9 +803,7 @@ def _filter_by_platform(releases: list[PyPI_Release], *, tags: frozenset[Platfor
         )
 
     filtered = [release for release in releases if compatible(release, include_sdist=False)]
-    if not filtered:
-        return [release for release in releases if compatible(release, include_sdist=True)]
-    return filtered
+    return filtered or [release for release in releases if compatible(release, include_sdist=True)]
 
 
 @beartype
@@ -895,8 +906,6 @@ def _get_version(name: Optional[str] = None, package_name=None, /, mod=None) ->
     version = getattr(mod, "version", version)
     if callable(version):
         version = version()
-    if isinstance(version, str):
-        return Version(version)
     return Version(version)
 
 
@@ -988,7 +997,7 @@ def _real_path(*, path: Path, _applied_decorators: dict, landmark) -> tuple[str,
             else:
                 frame = frame.f_back
         # a few more steps..
-        for x in _applied_decorators[landmark]:
+        for _ in _applied_decorators[landmark]:
             frame = frame.f_back
         try:
             # frame is in __call__ (or rather methdispatch), we need to step two frames back
@@ -1171,3 +1180,25 @@ def _check(x, y):
         return issubclass(y, x)
     except TypeError:  # (int => list[int]) or (list[int] => int) NOK
         return False
+
+
+# function to run code in a jupyter notebook sandbox
+def run_in_notebook(code):
+    """Run code in a jupyter notebook sandbox.
+
+    This function is useful to run code in a jupyter notebook
+    without having to run the notebook server.
+
+    Parameters
+    ----------
+    code : str
+        The code to run.
+
+    Returns
+    -------
+    None
+
+    """
+    from IPython.display import Javascript, display
+
+    display(Javascript(f"IPython.notebook.kernel.execute('{code}')"))
